"""
Code for searching collections of signatures.
"""

import csv
import numpy as np
from enum import Enum
from dataclasses import dataclass

from .minhash import MinHash
from .signature import SourmashSignature
from .sketchcomparison import FracMinHashComparison, NumMinHashComparison


def calc_threshold_from_bp(threshold_bp, scaled, query_size):
    """
    Convert threshold_bp (threshold in estimated bp) to
    fraction of query & minimum number of hashes needed.
    """
    threshold = 0.0
    n_threshold_hashes = 0

    if threshold_bp:
        if threshold_bp < 0:
            raise TypeError("threshold_bp must be non-negative")

        # if we have a threshold_bp of N, then that amounts to N/scaled
        # hashes:
        n_threshold_hashes = float(threshold_bp) / scaled

        # that then requires the following containment:
        threshold = n_threshold_hashes / query_size

        # is it too high to ever match?
        if threshold > 1.0:
            raise ValueError("requested threshold_bp is unattainable with this query")
    return threshold, n_threshold_hashes


class SearchType(Enum):
    JACCARD = 1
    CONTAINMENT = 2
    MAX_CONTAINMENT = 3


def make_jaccard_search_query(
    *, do_containment=False, do_max_containment=False, best_only=False, threshold=None
):
    """\
    Make a "flat" search object for Jaccard search & containment.
    """
    if do_containment and do_max_containment:
        raise TypeError("'do_containment' and 'do_max_containment' cannot both be True")

    # configure search - containment? ignore abundance? best only?
    search_cls = JaccardSearch
    if best_only:
        search_cls = JaccardSearchBestOnly

    if do_containment:
        search_obj = search_cls(SearchType.CONTAINMENT, threshold)
    elif do_max_containment:
        search_obj = search_cls(SearchType.MAX_CONTAINMENT, threshold)
    else:
        search_obj = search_cls(SearchType.JACCARD, threshold)

    return search_obj


def make_containment_query(query_mh, threshold_bp, *, best_only=True):
    "Make a search object for containment, with threshold_bp."
    if not query_mh:
        raise ValueError("query is empty!?")

    scaled = query_mh.scaled
    if not scaled:
        raise TypeError("query signature must be calculated with scaled")

    # are we setting a threshold?
    threshold, _ = calc_threshold_from_bp(threshold_bp, scaled, len(query_mh))

    if best_only:
        search_obj = JaccardSearchBestOnly(SearchType.CONTAINMENT, threshold=threshold)
    else:
        search_obj = JaccardSearch(SearchType.CONTAINMENT, threshold=threshold)

    return search_obj


class JaccardSearch:
    """
    A class used by Index classes for searching/gathering.
    """

    def __init__(self, search_type, threshold=None):
        "Constructor. Takes type of search, and optional threshold."
        score_fn = None
        require_scaled = False

        if search_type == SearchType.JACCARD:
            score_fn = self.score_jaccard
        elif search_type == SearchType.CONTAINMENT:
            score_fn = self.score_containment
            require_scaled = True
        elif search_type == SearchType.MAX_CONTAINMENT:
            score_fn = self.score_max_containment
            require_scaled = True
        self.score_fn = score_fn
        self.require_scaled = require_scaled

        if threshold is None:
            threshold = 0
        self.threshold = float(threshold)

    def check_is_compatible(self, sig):
        """
        Is this query compatible with this type of search? Raise TypeError
        if not.
        """
        if self.require_scaled:
            if not sig.minhash.scaled:
                raise TypeError("this search requires a scaled signature")

        if sig.minhash.track_abundance:
            raise TypeError("this search cannot be done with an abund signature")

    def passes(self, score):
        """Return True if this score meets or exceeds the threshold.

        Note: this can be used whenever a score or estimate is available
        (e.g. internal nodes on an SBT). `collect(...)`, below, decides
        whether a particular signature should be collected, and/or can
        update the threshold (used for BestOnly behavior).
        """
        if score and score >= self.threshold:
            return True
        return False

    def collect(self, score, match_sig):
        "Return True if this match should be collected."
        return True

    def score_jaccard(self, query_size, shared_size, subject_size, total_size):
        "Calculate Jaccard similarity."
        if total_size == 0:
            return 0
        return shared_size / total_size

    def score_containment(self, query_size, shared_size, subject_size, total_size):
        "Calculate Jaccard containment."
        if query_size == 0:
            return 0
        return shared_size / query_size

    def score_max_containment(self, query_size, shared_size, subject_size, total_size):
        "Calculate Jaccard max containment."
        min_denom = min(query_size, subject_size)
        if min_denom == 0:
            return 0
        return shared_size / min_denom


class JaccardSearchBestOnly(JaccardSearch):
    "A subclass of JaccardSearch that implements best-only."

    def collect(self, score, match):
        "Raise the threshold to the best match found so far."
        self.threshold = max(self.threshold, score)
        return True


@dataclass
class BaseResult:
    """
    Base class for sourmash search results.
    Since we need some additional info (scaled vs num minhashes) to
    properly initialize a SketchComparison, this class doesn't actually do
    anything other than define some functions needed by *Result classes.
    """

    query: SourmashSignature
    match: SourmashSignature
    filename: str = None
    ignore_abundance: bool = False  # optionally ignore abundances
    # need these for scaled result comparisons
    estimate_ani_ci: bool = False
    ani_confidence: float = 0.95
    threshold_bp: int = None
    cmp_scaled: int = None
    write_cols: list = None
    potential_false_negative: bool = False

    def init_result(self):
        self.mh1 = self.query.minhash
        self.mh2 = self.match.minhash

    def build_fracminhashcomparison(self):
        self.cmp = FracMinHashComparison(
            self.mh1,
            self.mh2,
            cmp_scaled=self.cmp_scaled,
            threshold_bp=self.threshold_bp,
            ignore_abundance=self.ignore_abundance,
            estimate_ani_ci=self.estimate_ani_ci,
            ani_confidence=self.ani_confidence,
        )
        self.cmp_scaled = self.cmp.cmp_scaled
        self.query_scaled = self.mh1.scaled
        self.match_scaled = self.mh2.scaled
        self.size_may_be_inaccurate = self.cmp.size_may_be_inaccurate

    def build_numminhashcomparison(self, cmp_num=None):
        self.cmp = NumMinHashComparison(
            self.mh1, self.mh2, cmp_num=cmp_num, ignore_abundance=self.ignore_abundance
        )
        self.cmp_num = self.cmp.cmp_num
        self.query_num = self.mh1.num
        self.match_num = self.mh2.num
        self.size_may_be_inaccurate = self.cmp.size_may_be_inaccurate

    def get_cmpinfo(self):
        # grab signature /minhash metadata
        # note, with so few columns written for current SearchResult, we could move these to PrefetchResult initialization
        # I've left here for now because I think at some point we want to output more info from search..
        self.ksize = self.mh1.ksize
        self.moltype = self.mh1.moltype
        self.query_name = self.query.name
        self.query_filename = self.query.filename
        self.query_md5 = self.query.md5sum()
        self.match_name = self.match.name
        self.match_filename = self.match.filename
        # sometimes filename is not set in sig (match_filename is None),
        # and `search` is able to pass in the filename.
        if self.filename is None and self.match_filename is not None:
            self.filename = self.match_filename
        self.match_md5 = self.match.md5sum()
        # set these from self.match_*
        self.md5 = self.match_md5
        self.name = self.match_name
        # could define in PrefetchResult instead, same reasoning as above
        self.query_abundance = self.mh1.track_abundance
        self.match_abundance = self.mh2.track_abundance
        self.query_n_hashes = len(self.mh1)
        self.match_n_hashes = len(self.mh2)

    @property
    def pass_threshold(self):
        return self.cmp.pass_threshold

    def shorten_md5(self, md5):
        return md5[:8]

    def to_write(self, columns=[]):
        # convert comparison attrs into a dictionary
        # that can be used by csv dictwriter
        info = {
            k: v for k, v in self.__dict__.items() if k in columns and v is not None
        }
        return info

    def init_dictwriter(self, csv_handle):
        # initialize the csv, return writer (do once)
        w = csv.DictWriter(csv_handle, fieldnames=self.write_cols)
        w.writeheader()
        return w

    def prep_result(self):
        # shorten or modify any columns that need modifying
        self.query_md5 = self.shorten_md5(self.query_md5)

    def write(self, w):
        # write result dictionary using csv dictwriter
        self.prep_result()
        w.writerow(self.to_write(columns=w.fieldnames))

    @property
    def resultdict(self):
        # instead of writing, just return dictionary of what we want to write
        self.prep_result()
        return self.to_write(columns=self.write_cols)


@dataclass
class SearchResult(BaseResult):
    """
    SearchResult class supports 'sourmash search' operations.
    """

    similarity: float = None
    cmp_num: int = None
    searchtype: SearchType = None

    # columns for standard SearchResult output
    search_write_cols = [
        "similarity",
        "md5",
        "filename",
        "name",  # here we use 'filename'
        "query_filename",
        "query_name",
        "query_md5",
        "ani",
    ]

    ci_cols = ["ani_low", "ani_high"]

    search_write_cols_ci = search_write_cols + ci_cols

    def init_sigcomparison(self):
        self.init_result()
        if any([self.mh1.scaled, self.mh2.scaled]):
            self.build_fracminhashcomparison()
        elif any([self.mh1.num, self.mh2.num]):
            self.build_numminhashcomparison(cmp_num=self.cmp_num)
        self.get_cmpinfo()  # grab comparison metadata

    def __post_init__(self):
        self.init_sigcomparison()  # build sketch comparison
        self.check_similarity()
        if self.cmp_scaled is not None and self.searchtype is not None:
            self.estimate_search_ani()
        # define columns we want to write
        self.write_cols = self.search_write_cols
        if self.estimate_ani_ci:
            self.write_cols = self.search_write_cols_ci

    def check_similarity(self):
        # for now, require similarity for SearchResult
        # future: consider returning SearchResult *during* search, and passing SearchType in.
        # then allow similarity to be calculated here according to SearchType.
        if self.similarity is None:
            raise ValueError("Error: Must provide 'similarity' for SearchResult.")

    def estimate_search_ani(self):
        # future: could estimate ANI from abund searches if we want (use query containment?)
        if self.cmp_scaled is None:
            raise TypeError("Error: ANI can only be estimated from scaled signatures.")
        if self.searchtype == SearchType.CONTAINMENT:
            self.cmp.estimate_ani_from_mh1_containment_in_mh2(
                containment=self.similarity
            )
            self.ani = self.cmp.ani_from_mh1_containment_in_mh2
            if self.estimate_ani_ci:
                self.ani_low = self.cmp.ani_from_mh1_containment_in_mh2_low
                self.ani_high = self.cmp.ani_from_mh1_containment_in_mh2_high
        elif self.searchtype == SearchType.MAX_CONTAINMENT:
            self.cmp.estimate_max_containment_ani()
            self.ani = self.cmp.max_containment_ani
            if self.estimate_ani_ci:
                self.ani_low = self.cmp.max_containment_ani_low
                self.ani_high = self.cmp.max_containment_ani_high
        elif self.searchtype == SearchType.JACCARD:
            self.cmp.estimate_jaccard_ani(jaccard=self.similarity)
            self.jaccard_ani_untrustworthy = self.cmp.jaccard_ani_untrustworthy
            self.ani = self.cmp.jaccard_ani
        # this can be set from any of the above
        self.potential_false_negative = self.cmp.potential_false_negative


@dataclass
class PrefetchResult(BaseResult):
    """
    PrefetchResult class supports 'sourmash prefetch' operations.
    """

    # current prefetch columns
    prefetch_write_cols = [
        "intersect_bp",
        "jaccard",
        "max_containment",
        "f_query_match",
        "f_match_query",
        "match_filename",
        "match_name",  # here we use 'match_filename'
        "match_md5",
        "match_bp",
        "query_filename",
        "query_name",
        "query_md5",
        "query_bp",
        "ksize",
        "moltype",
        "scaled",
        "query_n_hashes",
        "query_abundance",
        "query_containment_ani",
        "match_containment_ani",
        "average_containment_ani",
        "max_containment_ani",
        "potential_false_negative",
    ]  #'match_abundance'

    ci_cols = [
        "query_containment_ani_low",
        "query_containment_ani_high",
        "match_containment_ani_low",
        "match_containment_ani_high",
    ]

    prefetch_write_cols_ci = prefetch_write_cols + ci_cols

    def init_sigcomparison(self):
        # shared prefetch/gather initialization
        self.init_result()
        if all([self.mh1.scaled, self.mh2.scaled]):
            self.build_fracminhashcomparison()
        else:
            raise TypeError(
                "Error: prefetch and gather results must be between scaled signatures."
            )
        self.get_cmpinfo()  # grab comparison metadata
        self.intersect_bp = self.cmp.total_unique_intersect_hashes
        self.max_containment = self.cmp.max_containment
        self.query_bp = self.mh1.unique_dataset_hashes
        self.match_bp = self.mh2.unique_dataset_hashes
        self.threshold = self.threshold_bp
        self.estimate_containment_ani()

    def estimate_containment_ani(self):
        self.cmp.estimate_all_containment_ani()
        self.query_containment_ani = self.cmp.ani_from_mh1_containment_in_mh2
        self.match_containment_ani = self.cmp.ani_from_mh2_containment_in_mh1
        self.average_containment_ani = self.cmp.avg_containment_ani
        self.max_containment_ani = self.cmp.max_containment_ani
        self.potential_false_negative = self.cmp.potential_false_negative
        if self.estimate_ani_ci:
            self.handle_ani_ci()

    def handle_ani_ci(self):
        self.query_containment_ani_low = self.cmp.ani_from_mh1_containment_in_mh2_low
        self.query_containment_ani_high = self.cmp.ani_from_mh1_containment_in_mh2_high
        self.match_containment_ani_low = self.cmp.ani_from_mh2_containment_in_mh1_low
        self.match_containment_ani_high = self.cmp.ani_from_mh2_containment_in_mh1_high

    def build_prefetch_result(self):
        # unique prefetch values
        self.jaccard = self.cmp.jaccard
        self.f_query_match = (
            self.cmp.mh2_containment_in_mh1
        )  # db_mh.contained_by(query_mh)
        self.f_match_query = (
            self.cmp.mh1_containment_in_mh2
        )  # query_mh.contained_by(db_mh)
        # set write columns for prefetch result
        self.write_cols = self.prefetch_write_cols
        if self.estimate_ani_ci:
            self.write_cols = self.prefetch_write_cols_ci

    def __post_init__(self):
        self.init_sigcomparison()
        self.build_prefetch_result()

    def prep_prefetch_result(self):
        # explicitly name so we can use this within GatherResult too
        self.scaled = self.cmp_scaled
        # in prefetch, we shorten all md5's
        self.query_md5 = self.shorten_md5(self.query_md5)
        self.md5 = self.shorten_md5(self.md5)
        self.match_md5 = self.shorten_md5(self.match_md5)

    def prep_result(self):
        # overwrite base prep_result
        self.prep_prefetch_result()

    @property
    def prefetchresultdict(self):
        # just return dictionary of what we want to write
        self.prep_prefetch_result()
        return self.to_write(columns=self.write_cols)


@dataclass
class GatherResult(PrefetchResult):
    gather_querymh: MinHash = None
    gather_result_rank: int = None
    orig_query_len: int = None
    orig_query_abunds: list = None
    sum_weighted_found: int = None
    total_weighted_hashes: int = None
    noident_len: int = 0

    gather_write_cols = [
        "intersect_bp",
        "f_orig_query",
        "f_match",
        "f_unique_to_query",
        "f_unique_weighted",
        "average_abund",
        "median_abund",
        "std_abund",
        "filename",
        "name",
        "md5",
        "f_match_orig",
        "unique_intersect_bp",
        "gather_result_rank",
        "remaining_bp",
        "query_filename",
        "query_name",
        "query_md5",
        "query_bp",
        "ksize",
        "moltype",
        "scaled",
        "query_n_hashes",
        "query_abundance",
        "query_containment_ani",
        "match_containment_ani",
        "average_containment_ani",
        "max_containment_ani",
        "potential_false_negative",
        "n_unique_weighted_found",
        "sum_weighted_found",
        "total_weighted_hashes",
    ]

    ci_cols = [
        "query_containment_ani_low",
        "query_containment_ani_high",
        "match_containment_ani_low",
        "match_containment_ani_high",
    ]

    gather_write_cols_ci = gather_write_cols + ci_cols

    def init_gathersketchcomparison(self):
        # compare remaining gather hashes with match. Force at cmp_scaled. Force match flatten(), bc we don't need abunds.
        self.gather_comparison = FracMinHashComparison(
            self.gather_querymh, self.match.minhash.flatten()
        )

    def check_gatherresult_input(self):
        # check we have what we need:
        if self.cmp_scaled is None:
            raise ValueError(
                "Error: must provide comparison scaled value ('cmp_scaled') for GatherResult"
            )
        if self.gather_querymh is None:
            raise ValueError(
                "Error: must provide current gather sketch (remaining hashes) for GatherResult"
            )
        if self.gather_result_rank is None:
            raise ValueError("Error: must provide 'gather_result_rank' to GatherResult")
        if not self.total_weighted_hashes:  # catch total_weighted_hashes = 0 as well
            raise ValueError(
                "Error: must provide sum of all abundances ('total_weighted_hashes') to GatherResult"
            )
        if not self.orig_query_abunds:
            raise ValueError(
                "Error: must provide original query abundances ('orig_query_abunds') to GatherResult"
            )

    def build_gather_result(self):
        # build gather-specific attributes

        # the 'query' that is passed into gather is all _matched_ hashes, after subtracting noident_mh
        # this affects estimation of original query information, and requires us to pass in orig_query_len and orig_query_abunds.
        # we also need to overwrite self.query_bp, self.query_n_hashes, and self.query_abundance
        # todo: find a better solution?
        self.query_bp = self.orig_query_len * self.query.minhash.scaled
        self.query_n_hashes = self.orig_query_len

        # calculate intersection with query hashes:
        self.unique_intersect_bp = self.gather_comparison.total_unique_intersect_hashes

        # calculate fraction of subject match with orig query
        self.f_match_orig = self.cmp.mh2_containment_in_mh1

        # calculate fractions wrt first denominator - genome size
        self.f_match = (
            self.gather_comparison.mh2_containment_in_mh1
        )  # unique match containment
        self.f_orig_query = len(self.cmp.intersect_mh) / self.orig_query_len
        assert (
            self.gather_comparison.intersect_mh.contained_by(
                self.gather_comparison.mh1_cmp
            )
            == 1.0
        )

        # calculate fractions wrt second denominator - metagenome size
        assert (
            self.gather_comparison.intersect_mh.contained_by(
                self.gather_comparison.mh2_cmp
            )
            == 1.0
        )
        self.f_unique_to_query = (
            len(self.gather_comparison.intersect_mh) / self.orig_query_len
        )

        # here, need to make sure to use the mh1_cmp (bc was downsampled to cmp_scaled)
        self.remaining_bp = (
            self.noident_len
            + self.gather_comparison.mh1_cmp.unique_dataset_hashes
            - self.gather_comparison.total_unique_intersect_hashes
        )

        # calculate stats on abundances, if desired.
        self.average_abund, self.median_abund, self.std_abund = None, None, None
        if not self.ignore_abundance:
            self.query_weighted_unique_intersection = (
                self.gather_comparison.weighted_intersection(
                    from_abundD=self.orig_query_abunds
                )
            )
            self.average_abund = self.query_weighted_unique_intersection.mean_abundance
            self.median_abund = self.query_weighted_unique_intersection.median_abundance
            self.std_abund = self.query_weighted_unique_intersection.std_abundance
            # 'query' will be flattened by default. reset track abundance if we have abunds
            self.query_abundance = (
                self.query_weighted_unique_intersection.track_abundance
            )
            # calculate scores weighted by abundances
            self.n_unique_weighted_found = (
                self.query_weighted_unique_intersection.sum_abundances
            )
            self.f_unique_weighted = (
                self.n_unique_weighted_found / self.total_weighted_hashes
            )
        else:
            self.f_unique_weighted = self.f_unique_to_query
            self.query_abundance = False

    def __post_init__(self):
        self.check_gatherresult_input()
        self.init_sigcomparison()  # initialize original sketch vs match sketch comparison (inherited from PrefetchResult)
        self.init_gathersketchcomparison()  # initialize remaining gather sketch vs match sketch comparison
        self.build_gather_result()  # build gather-specific attributes
        # set write columns for prefetch result
        self.write_cols = self.gather_write_cols
        if self.estimate_ani_ci:
            self.write_cols = self.gather_write_cols_ci

    def prep_gather_result(self):
        # for gather, we only shorten the query_md5
        self.scaled = self.cmp_scaled
        self.query_md5 = self.shorten_md5(self.query_md5)

    def prep_result(self):
        # overwrite base prep_result
        self.prep_gather_result()

    @property
    def gatherresultdict(self):
        # just return dictionary of what we want to write
        self.prep_gather_result()
        return self.to_write(columns=self.write_cols)

    @property
    def prefetchresultdict(self):
        # get all prefetch cols from gatherresult
        prefetch_cols = self.prefetch_write_cols
        if self.estimate_ani_ci:
            prefetch_cols = self.prefetch_write_cols_ci
        self.jaccard = self.cmp.jaccard
        self.f_query_match = (
            self.cmp.mh2_containment_in_mh1
        )  # db_mh.contained_by(query_mh)
        self.f_match_query = (
            self.cmp.mh1_containment_in_mh2
        )  # query_mh.contained_by(db_mh)
        self.prep_prefetch_result()
        return self.to_write(columns=prefetch_cols)


def format_bp(bp):
    "Pretty-print bp information."
    bp = float(bp)
    if bp < 500:
        return f"{bp:.0f} bp"
    elif bp <= 500e3:
        return f"{round(bp / 1e3, 1):.1f} kbp"
    elif bp < 500e6:
        return f"{round(bp / 1e6, 1):.1f} Mbp"
    elif bp < 500e9:
        return f"{round(bp / 1e9, 1):.1f} Gbp"
    return "???"


def search_databases_with_flat_query(query, databases, **kwargs):
    results = []
    found_md5 = set()

    for db in databases:
        search_iter = db.search(query, **kwargs)
        for score, match, filename in search_iter:
            md5 = match.md5sum()
            if md5 not in found_md5:
                results.append((score, match, filename))
                found_md5.add(md5)

    # sort results on similarity (reverse)
    results.sort(key=lambda x: -x[0])

    # redefine searchtype and pass in here
    # repetitive/not optimal - would it be better to produce SearchResult from db.search?
    estimate_ani_ci = False
    search_type = SearchType.JACCARD
    if kwargs.get("do_containment"):
        search_type = SearchType.CONTAINMENT
        if kwargs.get("estimate_ani_ci"):
            estimate_ani_ci = True
    elif kwargs.get("do_max_containment"):
        search_type = SearchType.MAX_CONTAINMENT
        if kwargs.get("estimate_ani_ci"):
            estimate_ani_ci = True

    x = []
    for score, match, filename in results:
        x.append(
            SearchResult(
                query,
                match,
                similarity=score,
                filename=filename,
                searchtype=search_type,
                estimate_ani_ci=estimate_ani_ci,
            )
        )
    return x


def search_databases_with_abund_query(query, databases, **kwargs):
    results = []
    found_md5 = set()

    if kwargs.get("do_containment") or kwargs.get("do_max_containment"):
        raise TypeError("containment searches cannot be done with abund sketches")

    for db in databases:
        search_iter = db.search_abund(
            query, **kwargs
        )  # could return SearchResult here instead of tuple?
        for score, match, filename in search_iter:
            md5 = match.md5sum()
            if md5 not in found_md5:
                results.append((score, match, filename))
                found_md5.add(md5)

    # sort results on similarity (reverse)
    results.sort(key=lambda x: -x[0])

    x = []
    for score, match, filename in results:
        x.append(SearchResult(query, match, similarity=score, filename=filename))
    return x


###
### gather code
###


def _find_best(counters, query, threshold_bp):
    """
    Search for the best containment, return precisely one match.
    """
    best_result = None
    best_intersect_mh = None

    # find the best score across multiple counters, without consuming
    for counter in counters:
        result = counter.peek(query.minhash, threshold_bp=threshold_bp)
        if result:
            (sr, intersect_mh) = result

            if best_result is None or sr.score > best_result.score:
                best_result = sr
                best_intersect_mh = intersect_mh

    if best_result:
        # remove the best result from each counter
        for counter in counters:
            counter.consume(best_intersect_mh)

        # and done!
        return best_result, best_intersect_mh
    return None, None


class GatherDatabases:
    "Iterator object for doing gather/min-set-cov."

    def __init__(
        self,
        query,
        counters,
        *,
        threshold_bp=0,
        ignore_abundance=False,
        noident_mh=None,
        ident_mh=None,
        estimate_ani_ci=False,
    ):
        # track original query information for later usage?
        track_abundance = query.minhash.track_abundance and not ignore_abundance
        self.orig_query = query
        self.orig_query_bp = query.minhash.unique_dataset_hashes
        self.orig_query_filename = query.filename
        self.orig_query_name = query.name
        self.orig_query_md5 = query.md5sum()[:8]

        # do we pay attention to abundances?
        query_mh = query.minhash
        query_hashes = query_mh.hashes
        if track_abundance:
            orig_query_abunds = query_hashes
        else:
            orig_query_abunds = {k: 1 for k in query_hashes}

        # adjust for not found...
        if noident_mh is None:  # create empty
            noident_mh = query_mh.copy_and_clear()
        self.noident_mh = noident_mh.to_frozen()

        if ident_mh is None:
            query_mh = query_mh.to_mutable()
            query_mh.remove_many(noident_mh)
        else:
            query_mh = ident_mh.to_mutable()

        orig_query_mh = query_mh.flatten()

        # query.minhash will be assigned to repeatedly in gather; make mutable.
        query = query.to_mutable()
        query.minhash = orig_query_mh

        cmp_scaled = query.minhash.scaled  # initialize with resolution of query

        self.result_n = 0
        self.query = query
        self.counters = counters
        self.threshold_bp = threshold_bp

        self.track_abundance = track_abundance
        self.orig_query_mh = orig_query_mh
        self.orig_query_abunds = orig_query_abunds

        self.cmp_scaled = 0  # initialize with something very low!
        self._update_scaled(cmp_scaled)

        self.estimate_ani_ci = (
            estimate_ani_ci  # by default, do not report ANI confidence intervals
        )

    def _update_scaled(self, scaled):
        max_scaled = max(self.cmp_scaled, scaled)
        if self.cmp_scaled != max_scaled:
            self.cmp_scaled = max_scaled

            # CTB note: this can be expensive
            self.orig_query_mh = self.orig_query_mh.downsample(scaled=scaled)
            self.noident_mh = self.noident_mh.downsample(scaled=scaled)

            # NOTE: orig_query_abunds can be used w/o downsampling
            orig_query_abunds = self.orig_query_abunds
            self.noident_query_sum_abunds = sum(
                orig_query_abunds[k] for k in self.noident_mh.hashes
            )
            self.total_weighted_hashes = sum(
                orig_query_abunds[k] for k in self.orig_query_mh.hashes
            )
            self.total_weighted_hashes += self.noident_query_sum_abunds

        if max_scaled != scaled:
            return max_scaled
        return max_scaled

    @property
    def scaled(self):
        return self.cmp_scaled

    def __iter__(self):
        return self

    def __next__(self):
        query = self.query
        if not self.query.minhash:
            raise StopIteration

        # may be changed:
        counters = self.counters

        # will not be changed::
        threshold_bp = self.threshold_bp
        orig_query_abunds = self.orig_query_abunds

        # find the best match!
        best_result, intersect_mh = _find_best(counters, query, threshold_bp)

        if not best_result:  # no matches at all for this cutoff!
            raise StopIteration

        best_match = best_result.signature
        filename = best_result.location

        # Is the best match computed with scaled? Die if not.
        match_scaled = best_match.minhash.scaled
        assert match_scaled

        # pick the highest scaled / lowest resolution.
        scaled = self._update_scaled(match_scaled)
        # CTB note: this means that if a high scaled/low res signature is
        # found early on, resolution will be low from then on.

        # retrieve various saved things, after potential downsampling
        orig_query_mh = self.orig_query_mh
        total_weighted_hashes = self.total_weighted_hashes
        noident_mh = self.noident_mh
        orig_query_len = len(orig_query_mh) + len(noident_mh)

        # eliminate hashes under this new resolution.
        query_mh = query.minhash.downsample(scaled=scaled)
        found_mh = best_match.minhash.downsample(scaled=scaled).flatten()

        # construct a new query, subtracting hashes found in previous one.
        new_query_mh = query_mh.to_mutable()
        new_query_mh.remove_many(found_mh)
        new_query = SourmashSignature(new_query_mh)

        # compute weighted information for remaining query hashes
        query_hashes = set(new_query_mh.hashes)
        n_weighted_missed = sum(orig_query_abunds[k] for k in query_hashes)
        n_weighted_missed += self.noident_query_sum_abunds
        sum_weighted_found = total_weighted_hashes - n_weighted_missed

        # build a GatherResult
        result = GatherResult(
            self.orig_query,
            best_match,
            cmp_scaled=scaled,
            filename=filename,
            gather_result_rank=self.result_n,
            gather_querymh=query.minhash,
            ignore_abundance=not self.track_abundance,
            threshold_bp=threshold_bp,
            orig_query_len=orig_query_len,
            orig_query_abunds=self.orig_query_abunds,
            estimate_ani_ci=self.estimate_ani_ci,
            sum_weighted_found=sum_weighted_found,
            total_weighted_hashes=total_weighted_hashes,
            noident_len=len(self.noident_mh) * self.noident_mh.scaled,
        )

        self.result_n += 1
        self.query = new_query

        return result


###
### prefetch code
###


def prefetch_database(query, database, threshold_bp, *, estimate_ani_ci=False):
    """
    Find all matches to `query_mh` >= `threshold_bp` in `database`.
    """
    scaled = query.minhash.scaled
    assert scaled
    # iterate over all signatures in database, find matches
    for result in database.prefetch(
        query, threshold_bp
    ):  # future: could return PrefetchResult directly here
        result = PrefetchResult(
            query,
            result.signature,
            threshold_bp=threshold_bp,
            estimate_ani_ci=estimate_ani_ci,
        )
        assert result.pass_threshold
        yield result



"""search a signature against other signatures"""

usage = """

The `search` subcommand searches a collection of signatures or SBTs
for matches to the query signature.  It can search for matches with
either high Jaccard similarity [1] or containment; the default is to
use Jaccard similarity, unless `--containment` is specified.
`-o/--output` will create a CSV file containing the matches.

`search` will load all of provided signatures into memory, which can
be slow and somewhat memory intensive for large collections.  You can
use `sourmash index` to create a Sequence Bloom Tree (SBT) that can be
quickly searched on disk; this is the same format in which we provide
GenBank and other databases.

Command line usage:
```
sourmash search query.sig [ list of signatures or SBTs ]
```

Example output:

```
49 matches; showing first 20:
similarity   match
----------   -----
 75.4%%      NZ_JMGW01000001.1 Escherichia coli 1-176-05_S4_C2 e117605...
 72.2%%      NZ_GG774190.1 Escherichia coli MS 196-1 Scfld2538, whole ...
 71.4%%      NZ_JMGU01000001.1 Escherichia coli 2-011-08_S3_C2 e201108...
 70.1%%      NZ_JHRU01000001.1 Escherichia coli strain 100854 100854_1...
 69.0%%      NZ_JH659569.1 Escherichia coli M919 supercont2.1, whole g...
...  
```

[1] https://en.wikipedia.org/wiki/Jaccard_index

When `--containment` is provided, the containment of the query in each
of the search signatures or databases is reported.

---
"""

from sourmash.cli.utils import (
    add_ksize_arg,
    add_moltype_args,
    add_picklist_args,
    add_scaled_arg,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("search", description=__doc__, usage=usage)
    subparser.add_argument("query", help="query signature")
    subparser.add_argument(
        "databases",
        nargs="+",
        help="signatures/SBTs to search",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debug information"
    )
    subparser.add_argument(
        "-t",
        "--threshold",
        metavar="T",
        default=0.08,
        type=float,
        help="minimum threshold for reporting matches; default=0.08",
    )
    subparser.add_argument(
        "--save-matches",
        metavar="FILE",
        help="output matching signatures to the specified file",
    )
    subparser.add_argument(
        "--best-only",
        action="store_true",
        help="report only the best match (with greater speed)",
    )
    subparser.add_argument(
        "-n",
        "--num-results",
        default=3,
        type=int,
        metavar="N",
        help="number of results to display to user; 0 to report all",
    )
    subparser.add_argument(
        "--containment",
        action="store_true",
        help="score based on containment rather than similarity",
    )
    subparser.add_argument(
        "--max-containment",
        action="store_true",
        help="score based on max containment rather than similarity",
    )
    subparser.add_argument(
        "--estimate-ani-ci",
        action="store_true",
        help="for containment searches, also output confidence intervals for ANI estimates",
    )
    subparser.add_argument(
        "--ignore-abundance",
        action="store_true",
        help="do NOT use k-mer abundances if present; note: has no effect if "
        "--containment or --max-containment is specified",
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output CSV containing matches to this file",
    )
    subparser.add_argument(
        "--md5", default=None, help="select the signature with this md5 as query"
    )
    subparser.add_argument(
        "--fail-on-empty-database",
        action="store_true",
        help="stop at databases that contain no compatible signatures",
    )
    subparser.add_argument(
        "--no-fail-on-empty-database",
        action="store_false",
        dest="fail_on_empty_database",
        help="continue past databases that contain no compatible signatures",
    )
    subparser.set_defaults(fail_on_empty_database=True)

    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)
    add_pattern_args(subparser)
    add_scaled_arg(subparser, 0)


def main(args):
    import sourmash

    return sourmash.commands.search(args)



"LCA and reverse index utilities."

from .lca_db import LCA_Database
from .lca_utils import (
    taxlist,
    zip_lineage,
    build_tree,
    find_lca,
    gather_assignments,
    display_lineage,
    count_lca_for_assignments,
)

from .command_index import index
from .command_classify import classify
from .command_summarize import summarize_main
from .command_rankinfo import rankinfo_main
from .__main__ import main



"""search a metagenome signature against dbs"""

usage = """

The `gather` subcommand selects the best reference genomes to use for
a metagenome analysis, by finding the smallest set of non-overlapping
matches to the query in a database.  This is specifically meant for
metagenome and genome bin analysis.  (See "Classifying Signatures" [1]
in the command line documentation for more information on the
different approaches that can be used here.)

If the input signature was created with `-p abund`, output
will be abundance weighted (unless `--ignore-abundances` is
specified).  `-o/--output` will create a CSV file containing the
matches.

`gather`, like `search`, will load all of provided signatures into
memory.  You can use `sourmash index` to create a Sequence Bloom Tree
(SBT) that can be quickly searched on disk; this is the same format in
which we provide GenBank and other databases.

Command line usage:
```
sourmash gather query.sig [ list of signatures or SBTs ]
```

Example output for an unweighted/noabund query:
```
overlap     p_query p_match
---------   ------- --------
1.4 Mbp      11.0%%  58.0%%     JANA01000001.1 Fusobacterium sp. OBRC...
1.0 Mbp       7.7%%  25.9%%     CP001957.1 Haloferax volcanii DS2 pla...
0.9 Mbp       7.4%%  11.8%%     BA000019.2 Nostoc sp. PCC 7120 DNA, c...
0.7 Mbp       5.9%%  23.0%%     FOVK01000036.1 Proteiniclasticum rumi...
0.7 Mbp       5.3%%  17.6%%     AE017285.1 Desulfovibrio vulgaris sub...
```

Example output for a weighted query:
```
overlap     p_query p_match avg_abund
---------   ------- ------- ---------
9.3 Mbp        0.8%%   97.5%%       6.7    NC_007951.1 Burkholderia xenovorans ...
7.3 Mbp        2.3%%   99.9%%      23.9    NC_003272.1 Nostoc sp. PCC 7120 DNA,...
7.0 Mbp        8.9%%  100.0%%      94.5    BX119912.1 Rhodopirellula baltica SH...
6.6 Mbp        1.4%%  100.0%%      16.3    NC_009972.1 Herpetosiphon aurantiacu...
...
```

The command line option `--threshold-bp` sets the threshold below
which matches are no longer reported; by default, this is set to
50kb. see the Appendix in Classifying Signatures [1] for details.

Note:

Use `sourmash gather` to classify a metagenome against a collection of
genomes with no (or incomplete) taxonomic information.  Use `sourmash
lca summarize` to classify a metagenome using a collection of genomes
with taxonomic information.

[1] https://sourmash.readthedocs.io/en/latest/classifying-signatures.html

---
"""

from sourmash.cli.utils import (
    add_ksize_arg,
    add_moltype_args,
    add_picklist_args,
    add_scaled_arg,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("gather", description=__doc__, usage=usage)
    subparser.add_argument("query", help="query signature")
    subparser.add_argument(
        "databases",
        nargs="+",
        help="signatures/SBTs to search",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument("-d", "--debug", action="store_true")
    subparser.add_argument(
        "-n",
        "--num-results",
        default=None,
        type=int,
        metavar="N",
        help="number of results to report (default: terminate at --threshold-bp)",
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output CSV containing matches to this file",
    )
    subparser.add_argument(
        "--save-matches",
        metavar="FILE",
        help="save gather matched signatures from the database to the specified file",
    )
    subparser.add_argument(
        "--save-prefetch",
        metavar="FILE",
        help="save all prefetch-matched signatures from the databases to the "
        "specified file or directory",
    )
    subparser.add_argument(
        "--save-prefetch-csv",
        metavar="FILE",
        help="save a csv with information from all prefetch-matched signatures "
        "to the specified file",
    )
    subparser.add_argument(
        "--threshold-bp",
        metavar="REAL",
        type=float,
        default=5e4,
        help="reporting threshold (in bp) for estimated overlap with remaining query (default=50kb)",
    )
    subparser.add_argument(
        "--output-unassigned",
        metavar="FILE",
        help="output unassigned portions of the query as a signature to the "
        "specified file",
    )
    subparser.add_argument(
        "--ignore-abundance",
        action="store_true",
        help="do NOT use k-mer abundances if present",
    )
    subparser.add_argument(
        "--md5", default=None, help="select the signature with this md5 as query"
    )
    subparser.add_argument(
        "--cache-size",
        default=0,
        type=int,
        metavar="N",
        help="number of internal SBT nodes to cache in memory (default: 0, cache all nodes)",
    )

    # advanced parameters
    subparser.add_argument(
        "--linear",
        dest="linear",
        action="store_true",
        help="force a low-memory but maybe slower database search",
    )
    subparser.add_argument(
        "--no-linear",
        dest="linear",
        action="store_false",
    )
    subparser.add_argument(
        "--no-prefetch",
        dest="prefetch",
        action="store_false",
        help="do not use prefetch before gather; see documentation",
    )
    subparser.add_argument(
        "--prefetch",
        dest="prefetch",
        action="store_true",
        help="use prefetch before gather; see documentation",
    )
    subparser.add_argument(
        "--estimate-ani-ci",
        action="store_true",
        help="also output confidence intervals for ANI estimates",
    )
    subparser.add_argument(
        "--fail-on-empty-database",
        action="store_true",
        help="stop at databases that contain no compatible signatures",
    )
    subparser.add_argument(
        "--no-fail-on-empty-database",
        action="store_false",
        dest="fail_on_empty_database",
        help="continue past databases that contain no compatible signatures",
    )
    subparser.set_defaults(fail_on_empty_database=True)
    subparser.add_argument(
        "--create-empty-results",
        action="store_true",
        help="create an empty results file even if no matches.",
    )

    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)
    add_pattern_args(subparser)
    add_scaled_arg(subparser, 0)


def main(args):
    import sourmash

    return sourmash.commands.gather(args)



"""create protein signature from DNA/RNA sequence"""

usage = """

    sourmash sketch translate data/*.fna.gz

The 'sketch translate' command reads in DNA sequences and outputs protein
sketches.

By default, 'sketch translate' uses the parameter string
'k=10,scaled=200,noabund'.

This corresponds to a DNA k-mer size of 30 (and an amino-acid k-mer size
of 10), a scaled factor of 200, and no abundance tracking of
k-mers. You can specify one or more parameter strings of your own with
-p, e.g. 'sourmash sketch translate -p k=11,noabund -p
k=12,scaled=100,abund'.

'sourmash sketch' takes input sequences in FASTA and FASTQ,
uncompressed or gz/bz2 compressed.

Please see the 'sketch' documentation for more details:
  https://sourmash.readthedocs.io/en/latest/sourmash-sketch.html
"""

from sourmash import command_sketch

assert command_sketch.DEFAULTS["protein"] == "k=10,scaled=200,noabund"

import sourmash
from sourmash.logging import notify, print_results, error


def subparser(subparsers):
    subparser = subparsers.add_parser("translate", usage=usage)
    subparser.add_argument(
        "--license",
        default="CC0",
        type=str,
        help="signature license. Currently only CC0 is supported.",
    )
    subparser.add_argument(
        "--check-sequence",
        action="store_true",
        help="complain if input sequence is invalid DNA",
    )
    subparser.add_argument(
        "-p",
        "--param-string",
        default=[],
        help="signature parameters to use.",
        action="append",
    )

    subparser.add_argument("filenames", nargs="*", help="file(s) of sequences")
    file_args = subparser.add_argument_group("File handling options")
    file_args.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="recompute signatures even if the file exists",
    )
    file_args.add_argument(
        "-o", "--output", help="output computed signatures to this file"
    )
    subparser.add_argument(
        "--from-file", help="a text file containing a list of sequence files to load"
    )
    file_args.add_argument(
        "--set-name",
        "--name",
        "--merge",
        dest="merge",
        type=str,
        default="",
        metavar="FILE",
        help="name the output sketch as specified; note, merges all input "
        "files while sketching",
    )
    file_args.add_argument(
        "--output-dir",
        "--outdir",
        help="output computed signatures to this directory",
    )
    file_args.add_argument(
        "--singleton",
        action="store_true",
        help="compute a signature for each sequence record individually",
    )
    file_args.add_argument(
        "--name-from-first",
        action="store_true",
        help="name the signature generated from each file after the first "
        "record in the file",
    )
    file_args.add_argument(
        "--randomize",
        action="store_true",
        help="shuffle the list of input filenames randomly",
    )
    file_args.add_argument(
        "--dayhoff",
        action="store_true",
        help="compute sketches using the dayhoff alphabet instead",
    )
    file_args.add_argument(
        "--hp",
        action="store_true",
        help="compute sketches using the dayhoff alphabet instead",
    )


def main(args):
    import sourmash.command_sketch

    return sourmash.command_sketch.translate(args)



"Tests for search.py code."

import pytest
import numpy as np
import sourmash_tst_utils as utils

from sourmash import search, SourmashSignature, MinHash, load_one_signature
from sourmash.search import (
    make_jaccard_search_query,
    make_containment_query,
    SearchResult,
    PrefetchResult,
    GatherResult,
)
from sourmash.index import LinearIndex


def test_make_jaccard_search_query():
    search_obj = make_jaccard_search_query(threshold=0)

    assert search_obj.score_fn == search_obj.score_jaccard
    assert not search_obj.require_scaled
    assert search_obj.threshold == 0


def test_make_jaccard_search_query_cont():
    search_obj = make_jaccard_search_query(do_containment=True, threshold=0)

    assert search_obj.score_fn == search_obj.score_containment
    assert search_obj.require_scaled
    assert search_obj.threshold == 0


def test_make_jaccard_search_query_max_cont():
    search_obj = make_jaccard_search_query(do_max_containment=True, threshold=0)

    assert search_obj.score_fn == search_obj.score_max_containment
    assert search_obj.require_scaled
    assert search_obj.threshold == 0


def test_make_jaccard_search_query_best_only():
    search_obj = make_jaccard_search_query(best_only=True)

    assert search_obj.score_fn == search_obj.score_jaccard
    assert not search_obj.require_scaled
    assert type(search_obj) is search.JaccardSearchBestOnly


def test_make_jaccard_search_query_no_threshold_none():
    search_obj = make_jaccard_search_query(threshold=None)

    assert search_obj.score_fn == search_obj.score_jaccard
    assert not search_obj.require_scaled
    assert search_obj.threshold == 0


def test_make_jaccard_search_query_cont_and_max_cont():
    with pytest.raises(TypeError) as exc:
        make_jaccard_search_query(do_containment=True, do_max_containment=True)

    assert (
        str(exc.value)
        == "'do_containment' and 'do_max_containment' cannot both be True"
    )


def test_cont_requires_scaled():
    search_obj = make_jaccard_search_query(do_containment=True)
    assert search_obj.require_scaled

    mh = MinHash(n=500, ksize=31)
    with pytest.raises(TypeError) as exc:
        search_obj.check_is_compatible(SourmashSignature(mh))
    assert str(exc.value) == "this search requires a scaled signature"


def test_search_requires_flat():
    search_obj = make_jaccard_search_query()

    mh = MinHash(n=500, ksize=31, track_abundance=True)
    with pytest.raises(TypeError) as exc:
        search_obj.check_is_compatible(SourmashSignature(mh))
    assert str(exc.value) == "this search cannot be done with an abund signature"


def test_score_jaccard_similarity():
    search_obj = make_jaccard_search_query()

    assert search_obj.score_fn(None, 100, None, 200) == 0.5


def test_score_jaccard_containment():
    search_obj = make_jaccard_search_query(do_containment=True)

    assert search_obj.score_fn(100, 50, None, 0) == 0.5


def test_score_jaccard_containment_zero_query_size():
    search_obj = make_jaccard_search_query(do_containment=True)

    assert search_obj.score_fn(0, 100, None, None) == 0


def test_score_jaccard_max_containment_1():
    search_obj = make_jaccard_search_query(do_max_containment=True)

    assert search_obj.score_fn(150, 75, 100, None) == 0.75


def test_score_jaccard_max_containment_2():
    search_obj = make_jaccard_search_query(do_max_containment=True)

    assert search_obj.score_fn(100, 75, 150, None) == 0.75


def test_score_jaccard_max_containment_zero_query_size():
    search_obj = make_jaccard_search_query(do_containment=True)

    assert search_obj.score_fn(0, 100, None, None) == 0


def test_collect():
    search_obj = make_jaccard_search_query(threshold=0)
    search_obj.collect(1.0, None)
    assert search_obj.threshold == 0


def test_collect_best_only():
    search_obj = make_jaccard_search_query(threshold=0, best_only=True)
    search_obj.collect(1.0, None)
    assert search_obj.threshold == 1.0


def test_make_containment_query():
    # test basic make_containment_query call
    mh = MinHash(n=0, ksize=31, scaled=1000)

    for i in range(100):
        mh.add_hash(i)

    search_obj = make_containment_query(mh, 5e4)

    assert search_obj.score_fn == search_obj.score_containment
    assert search_obj.require_scaled
    assert search_obj.threshold == 0.5


def test_make_containment_query_no_threshold():
    # test basic make_containment_query call
    mh = MinHash(n=0, ksize=31, scaled=1000)

    for i in range(100):
        mh.add_hash(i)

    search_obj = make_containment_query(mh, None)

    assert search_obj.score_fn == search_obj.score_containment
    assert search_obj.require_scaled
    assert search_obj.threshold == 0


def test_make_containment_query_num_minhash():
    # will fail on non-scaled minhash
    mh = MinHash(n=500, ksize=31)

    for i in range(100):
        mh.add_hash(i)

    with pytest.raises(TypeError) as exc:
        make_containment_query(mh, 5e4)

    assert str(exc.value) == "query signature must be calculated with scaled"


def test_make_containment_query_empty_minhash():
    # will fail on non-scaled minhash
    mh = MinHash(n=0, ksize=31, scaled=1000)

    for i in range(100):
        mh.add_hash(i)

    with pytest.raises(TypeError) as exc:
        make_containment_query(mh, -1)

    assert str(exc.value) == "threshold_bp must be non-negative"


def test_make_containment_query_high_threshold():
    # will fail on non-scaled minhash
    mh = MinHash(n=0, ksize=31, scaled=1000)

    for i in range(100):
        mh.add_hash(i)

    # effective threshold > 1; raise ValueError
    with pytest.raises(ValueError):
        make_containment_query(mh, 200000)


class FakeIndex(LinearIndex):
    _signatures = []
    filename = "something_or_other"

    def __init__(self, validator_fn):
        self.validator = validator_fn

    def find(self, search_fn, query, *args, **kwargs):
        if self.validator:
            self.validator(search_fn, query, args, kwargs)
        else:
            assert 0, "what are we even doing here?"
        return []


def test_index_search_passthru():
    # check that kwargs are passed through from 'search' to 'find'
    query = None

    def validate_kwarg_passthru(search_fn, query, args, kwargs):
        assert "this_kw_arg" in kwargs
        assert kwargs["this_kw_arg"] == 5

    idx = FakeIndex(validate_kwarg_passthru)

    idx.search(query, threshold=0.0, this_kw_arg=5)


def test_index_containment_passthru():
    # check that kwargs are passed through from 'search' to 'find'
    query = None

    def validate_kwarg_passthru(search_fn, query, args, kwargs):
        assert "this_kw_arg" in kwargs
        assert kwargs["this_kw_arg"] == 5

    idx = FakeIndex(validate_kwarg_passthru)

    idx.search(query, threshold=0.0, this_kw_arg=5)


def test_search_with_abund_query():
    mh = MinHash(n=0, ksize=31, scaled=1, track_abundance=True)
    query = SourmashSignature(mh)

    with pytest.raises(TypeError):
        search.search_databases_with_abund_query(
            query, [], threshold=0, do_containment=True
        )

    with pytest.raises(TypeError):
        search.search_databases_with_abund_query(
            query, [], threshold=0, do_max_containment=True
        )


def test_scaledSearchResult():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    ss4763 = ss4763.to_mutable()
    ss4763.filename = ss4763_file

    scaled = ss47.minhash.scaled

    res = SearchResult(
        ss47, ss4763, cmp_scaled=scaled, similarity=ss47.contained_by(ss4763)
    )

    assert res.query_name == ss47.name
    assert res.match_name == ss4763.name
    assert res.query_scaled == ss47.minhash.scaled == 1000
    assert res.match_scaled == ss4763.minhash.scaled == 1000
    assert res.cmp_scaled == 1000
    assert res.query_abundance == ss47.minhash.track_abundance
    assert res.match_abundance == ss4763.minhash.track_abundance
    #    assert res.query_bp == len(ss47.minhash) * scaled
    #    assert res.match_bp == len(ss4763.minhash) * scaled
    assert res.ksize == 31
    assert res.moltype == "DNA"
    assert res.query_filename == "47.fa"
    assert res.match_filename == ss4763_file
    assert res.query_md5 == ss47.md5sum()
    assert res.match_md5 == ss4763.md5sum()
    #   assert res.query_n_hashes == len(ss47.minhash)
    #   assert res.match_n_hashes == len(ss4763.minhash)
    assert res.md5 == ss4763.md5sum()
    assert res.name == ss4763.name
    assert res.filename == ss4763.filename
    queryc_ani = ss47.containment_ani(ss4763)
    matchc_ani = ss4763.containment_ani(ss47)
    # check that we _can_ get avg_containment_ani
    assert res.cmp.avg_containment_ani == np.mean([queryc_ani.ani, matchc_ani.ani])


def test_numSearchResult():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("num/47.fa.sig")
    ss63_file = utils.get_test_data("num/63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss63 = load_one_signature(ss63_file, ksize=31, select_moltype="dna")
    ss63 = ss63.to_mutable()
    ss63.filename = ss63_file

    assert ss47.minhash.num and ss63.minhash.num

    res = SearchResult(ss47, ss63, similarity=ss47.jaccard(ss63))
    print(res.cmp_num)
    assert res.mh1.num
    assert res.cmp.cmp_num == 500
    assert res.query_name == ss47.name
    assert res.match_name == ss63.name
    assert res.query_num == ss47.minhash.num == 500
    assert res.match_num == ss63.minhash.num == 500
    assert res.query_abundance == ss47.minhash.track_abundance
    assert res.match_abundance == ss63.minhash.track_abundance
    assert res.ksize == 31
    assert res.moltype == "DNA"
    assert res.query_filename == "47.fa"
    assert res.match_filename == ss63_file
    assert res.query_md5 == ss47.md5sum()
    assert res.match_md5 == ss63.md5sum()
    assert res.md5 == ss63.md5sum()
    assert res.name == ss63.name
    assert res.filename == ss63.filename

    # check that we can't get ani
    with pytest.raises(TypeError) as exc:
        res.estimate_search_ani()
    assert ("ANI can only be estimated from scaled signatures.") in str(exc)

    # get result as dictionary (of just items to write)
    resD = res.resultdict
    assert resD["filename"] == res.filename
    assert resD["name"] == res.name
    assert resD["similarity"] == res.similarity


def test_SearchResult_incompatible_sigs():
    ss47_file = utils.get_test_data("num/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    with pytest.raises(TypeError) as exc:
        SearchResult(ss47, ss4763, similarity=10)
    print(str(exc))
    assert "Error: Both sketches must be 'num' or 'scaled'." in str(exc)


def test_SearchResult_notsigs():
    ss47_file = utils.get_test_data("num/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")

    with pytest.raises(AttributeError) as exc:
        SearchResult(ss47_file, ss4763_file, similarity=10)
    print(str(exc))
    assert "'str' object has no attribute 'minhash'" in str(exc)


def test_SearchResult_no_similarity():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    with pytest.raises(ValueError) as exc:
        SearchResult(ss47, ss4763)
    print(str(exc))
    assert "Error: Must provide 'similarity' for SearchResult." in str(exc)


def test_PrefetchResult():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    ss4763 = ss4763.to_mutable()
    ss4763.filename = ss4763_file

    scaled = ss47.minhash.scaled

    intersect_mh = ss47.minhash.intersection(ss4763.minhash)
    intersect_bp = len(intersect_mh) * scaled
    jaccard = ss4763.jaccard(ss47)
    max_containment = ss4763.max_containment(ss47)
    f_match_query = ss47.contained_by(ss4763)
    f_query_match = ss4763.contained_by(ss47)
    queryc_ani = ss47.containment_ani(ss4763)
    matchc_ani = ss4763.containment_ani(ss47)

    res = PrefetchResult(ss47, ss4763, cmp_scaled=scaled)

    assert res.query_name == ss47.name
    assert res.match_name == ss4763.name
    assert res.query_scaled == ss47.minhash.scaled == 1000
    assert res.match_scaled == ss4763.minhash.scaled == 1000
    assert res.cmp_scaled == 1000
    assert res.query_abundance == ss47.minhash.track_abundance
    assert res.match_abundance == ss4763.minhash.track_abundance
    assert res.query_bp == len(ss47.minhash) * scaled
    assert res.match_bp == len(ss4763.minhash) * scaled
    assert res.ksize == 31
    assert res.moltype == "DNA"
    assert res.query_filename == "47.fa"
    assert res.match_filename == ss4763_file
    assert res.query_md5 == ss47.md5sum()
    assert res.match_md5 == ss4763.md5sum()
    assert res.query_n_hashes == len(ss47.minhash)
    assert res.match_n_hashes == len(ss4763.minhash)
    assert res.md5 == ss4763.md5sum()
    assert res.name == ss4763.name
    assert res.intersect_bp == intersect_bp
    assert res.jaccard == jaccard
    assert res.max_containment == max_containment
    assert res.f_query_match == f_query_match
    assert res.f_match_query == f_match_query

    # check ani
    assert res.query_containment_ani == queryc_ani.ani
    print(queryc_ani.ani)
    print(matchc_ani.ani)
    assert res.match_containment_ani == matchc_ani.ani
    assert res.max_containment_ani == max(queryc_ani.ani, matchc_ani.ani)
    assert res.average_containment_ani == np.mean([queryc_ani.ani, matchc_ani.ani])
    assert res.potential_false_negative == False


def test_PrefetchResult_incompatible_sigs():
    ss47_file = utils.get_test_data("num/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    with pytest.raises(TypeError) as exc:
        PrefetchResult(ss47, ss4763)
    print(str(exc))
    assert (
        "Error: prefetch and gather results must be between scaled signatures."
        in str(exc)
    )


def test_GatherResult():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("track_abund/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    ss4763 = ss4763.to_mutable()
    ss4763.filename = ss4763_file

    scaled = ss47.minhash.scaled

    intersect_mh = ss47.minhash.flatten().intersection(ss4763.minhash)
    remaining_mh = ss4763.minhash.to_mutable()
    remaining_mh.remove_many(intersect_mh)

    intersect_bp = len(intersect_mh) * scaled
    max_containment = ss4763.max_containment(ss47)
    ss47.contained_by(ss4763)
    orig_query_abunds = ss47.minhash.hashes
    queryc_ani = ss47.containment_ani(ss4763)
    matchc_ani = ss4763.containment_ani(ss47)

    # make some fake vals to check
    gather_result_rank = 1
    sum_abunds = 1000

    res = GatherResult(
        ss47,
        ss4763,
        cmp_scaled=scaled,
        gather_querymh=remaining_mh,
        gather_result_rank=gather_result_rank,
        total_weighted_hashes=sum_abunds,
        orig_query_len=len(ss47.minhash),
        orig_query_abunds=orig_query_abunds,
    )

    assert res.query_name == ss47.name
    assert res.match_name == ss4763.name
    assert res.query_scaled == ss47.minhash.scaled == 1000
    assert res.match_scaled == ss4763.minhash.scaled == 1000
    assert res.cmp_scaled == 1000
    assert res.query_abundance == ss47.minhash.track_abundance
    assert res.match_abundance == ss4763.minhash.track_abundance
    assert res.query_bp == len(ss47.minhash) * scaled
    assert res.match_bp == len(ss4763.minhash) * scaled
    assert res.ksize == 31
    assert res.moltype == "DNA"
    assert res.query_filename == "podar-ref/47.fa"
    assert res.match_filename == ss4763_file
    assert res.query_md5 == ss47.md5sum()
    assert res.match_md5 == ss4763.md5sum()
    assert res.query_n_hashes == len(ss47.minhash)
    assert res.match_n_hashes == len(ss4763.minhash)
    assert res.query_bp == ss47.minhash.unique_dataset_hashes
    assert res.match_bp == ss4763.minhash.unique_dataset_hashes
    assert res.md5 == ss4763.md5sum()
    assert res.name == ss4763.name
    assert res.match_filename == ss4763.filename
    assert res.intersect_bp == intersect_bp
    assert res.max_containment == max_containment

    # check that we can write prefetch result directly from gather
    pf = PrefetchResult(ss47, ss4763, cmp_scaled=scaled)
    assert pf.prefetchresultdict == res.prefetchresultdict

    # check ani
    assert res.query_containment_ani == queryc_ani.ani
    print(queryc_ani.ani)
    print(matchc_ani.ani)
    assert res.match_containment_ani == matchc_ani.ani
    assert res.max_containment_ani == max(queryc_ani.ani, matchc_ani.ani)
    assert res.average_containment_ani == np.mean([queryc_ani.ani, matchc_ani.ani])
    assert res.potential_false_negative == False

    # get write dict version of GatherResult
    resD = res.gatherresultdict
    assert resD["intersect_bp"] == res.intersect_bp


def test_GatherResult_ci():
    # check that values get stored/calculated correctly
    ss47_file = utils.get_test_data("track_abund/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")

    ss4763 = ss4763.to_mutable()
    ss4763.filename = ss4763_file

    scaled = ss47.minhash.scaled

    intersect_mh = ss47.minhash.flatten().intersection(ss4763.minhash)
    remaining_mh = ss4763.minhash.to_mutable()
    remaining_mh.remove_many(intersect_mh)

    orig_query_abunds = ss47.minhash.hashes
    queryc_ani = ss47.containment_ani(ss4763, estimate_ci=True)
    matchc_ani = ss4763.containment_ani(ss47, estimate_ci=True)

    # make some fake vals to check
    gather_result_rank = 1
    sum_abunds = 1000

    res = GatherResult(
        ss47,
        ss4763,
        cmp_scaled=scaled,
        gather_querymh=remaining_mh,
        gather_result_rank=gather_result_rank,
        total_weighted_hashes=sum_abunds,
        orig_query_len=len(ss47.minhash),
        orig_query_abunds=orig_query_abunds,
        estimate_ani_ci=True,
    )

    # check that we can write prefetch result directly from gather
    pf = PrefetchResult(ss47, ss4763, cmp_scaled=scaled, estimate_ani_ci=True)
    assert pf.prefetchresultdict == res.prefetchresultdict

    # check ani
    assert res.query_containment_ani == queryc_ani.ani
    print(queryc_ani.ani)
    print(matchc_ani.ani)
    assert res.match_containment_ani == matchc_ani.ani
    assert res.match_containment_ani_low == matchc_ani.ani_low
    assert res.match_containment_ani_high == matchc_ani.ani_high
    assert res.max_containment_ani == max(queryc_ani.ani, matchc_ani.ani)
    assert res.average_containment_ani == np.mean([queryc_ani.ani, matchc_ani.ani])
    assert res.potential_false_negative == False

    # get write dict version of GatherResult
    resD = res.gatherresultdict
    assert resD["intersect_bp"] == res.intersect_bp
    assert resD["match_containment_ani_low"] == res.match_containment_ani_low


def test_GatherResult_incompatible_sigs():
    ss47_file = utils.get_test_data("num/47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = ss47.minhash.hashes

    with pytest.raises(TypeError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: prefetch and gather results must be between scaled signatures."
        in str(exc)
    )


def test_GatherResult_incomplete_input_cmpscaled():
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = ss47.minhash.hashes

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=None,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide comparison scaled value ('cmp_scaled') for GatherResult"
        in str(exc)
    )


def test_GatherResult_incomplete_input_gathermh():
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = ss47.minhash.hashes

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=None,
            gather_result_rank=1,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide current gather sketch (remaining hashes) for GatherResult"
        in str(exc)
    )


def test_GatherResult_incomplete_input_gather_result_rank():
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = ss47.minhash.hashes

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=ss47.minhash,
            gather_result_rank=None,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert "Error: must provide 'gather_result_rank' to GatherResult" in str(exc)


def test_GatherResult_incomplete_input_total_weighted_hashes():
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = ss47.minhash.hashes

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=None,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide sum of all abundances ('total_weighted_hashes') to GatherResult"
        in str(exc)
    )

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=0,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide sum of all abundances ('total_weighted_hashes') to GatherResult"
        in str(exc)
    )


def test_GatherResult_incomplete_input_orig_query_abunds():
    ss47_file = utils.get_test_data("47.fa.sig")
    ss4763_file = utils.get_test_data("47+63.fa.sig")
    ss47 = load_one_signature(ss47_file, ksize=31, select_moltype="dna")
    ss4763 = load_one_signature(ss4763_file, ksize=31, select_moltype="dna")
    orig_query_abunds = None

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide original query abundances ('orig_query_abunds') to GatherResult"
        in str(exc)
    )

    orig_query_abunds = {}

    with pytest.raises(ValueError) as exc:
        GatherResult(
            ss47,
            ss4763,
            cmp_scaled=1000,
            gather_querymh=ss47.minhash,
            gather_result_rank=1,
            total_weighted_hashes=1,
            orig_query_len=len(ss47.minhash),
            orig_query_abunds=orig_query_abunds,
        )
    print(str(exc))
    assert (
        "Error: must provide original query abundances ('orig_query_abunds') to GatherResult"
        in str(exc)
    )



"""search a signature against dbs, find all overlaps"""

from sourmash.cli.utils import (
    add_ksize_arg,
    add_moltype_args,
    add_picklist_args,
    add_scaled_arg,
    add_pattern_args,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("prefetch")
    subparser.add_argument("query", help="query signature")
    subparser.add_argument(
        "databases",
        nargs="*",
        help="one or more databases to search",
    )
    subparser.add_argument(
        "--db-from-file",
        default=None,
        help="list of paths containing signatures to search",
    )
    subparser.add_argument(
        "--linear",
        action="store_true",
        help="force linear traversal of indexes to minimize loading time and memory use",
    )
    subparser.add_argument(
        "--no-linear",
        dest="linear",
        action="store_false",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument("-d", "--debug", action="store_true")
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output CSV containing matches to this file",
    )
    subparser.add_argument(
        "--save-matches",
        metavar="FILE",
        help="save all matching signatures from the databases to the "
        "specified file or directory",
    )
    subparser.add_argument(
        "--threshold-bp",
        metavar="REAL",
        type=float,
        default=5e4,
        help="reporting threshold (in bp) for estimated overlap with remaining query hashes (default=50kb)",
    )
    subparser.add_argument(
        "--save-unmatched-hashes",
        metavar="FILE",
        help="output unmatched query hashes as a signature to the specified file",
    )
    subparser.add_argument(
        "--save-matching-hashes",
        metavar="FILE",
        help="output matching query hashes as a signature to the specified file",
    )
    subparser.add_argument(
        "--md5", default=None, help="select the signature with this md5 as query"
    )
    subparser.add_argument(
        "--estimate-ani-ci",
        action="store_true",
        help="also output confidence intervals for ANI estimates",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)
    add_pattern_args(subparser)
    add_scaled_arg(subparser, 0)


def main(args):
    import sourmash

    return sourmash.commands.prefetch(args)



"""database rank info"""


def subparser(subparsers):
    subparser = subparsers.add_parser("rankinfo")
    subparser.add_argument("db", nargs="+")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="output debugging output"
    )
    subparser.add_argument("--scaled", metavar="FLOAT", type=float)
    subparser.add_argument(
        "--minimum-num",
        type=int,
        default=0,
        help="Minimum number of different lineages a k-mer must be in to be counted",
    )


def main(args):
    import sourmash

    return sourmash.lca.command_rankinfo.rankinfo_main(args)



"""intersect two or more signatures"""

usage = """

### `sourmash signature intersect` - intersect two (or more) signatures

Output the intersection of the hash values in multiple signature files.

For example,


sourmash signature intersect file1.sig file2.sig file3.sig -o intersect.sig

will output the intersection of all the hashes in those three files to
`intersect.sig`.

The `intersect` command flattens all signatures, i.e. the abundances
in any signatures will be ignored and the output signature will have
`track_abundance` turned off. See `sourmash signature flatten` for more details.

Note: `intersect` only creates one output file, with one signature in it.

"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("intersect", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        default="-",
        help="output signature to this file (default stdout)",
    )
    subparser.add_argument(
        "-A",
        "--abundances-from",
        metavar="FILE",
        help="intersect with & take abundances from this signature",
    )
    subparser.add_argument("--set-name", help="set name for output signature")
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.intersect(args)



"""classify a stream of sequences"""

from sourmash.cli.utils import add_ksize_arg, add_moltype_args, add_num_arg


def subparser(subparsers):
    subparser = subparsers.add_parser("watch")
    subparser.add_argument("sbt_name", help="name of SBT to search")
    subparser.add_argument("inp_file", nargs="?", default="/dev/stdin")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o", "--output", help="save signature generated from data here"
    )
    subparser.add_argument(
        "--threshold",
        metavar="T",
        default=0.05,
        type=float,
        help="minimum threshold for matches (default=0.05)",
    )
    subparser.add_argument(
        "--input-is-protein",
        action="store_true",
        help="Consume protein sequences - no translation needed",
    )
    add_moltype_args(subparser)
    subparser.add_argument(
        "--name", type=str, default="stdin", help="name to use for generated signature"
    )
    add_ksize_arg(subparser)
    add_num_arg(subparser, 500)


def main(args):
    import sourmash

    return sourmash.commands.watch(args)



"""subtract one or more signatures"""

usage = """

### `sourmash signature subtract` - subtract other signatures from a signature

Subtract all of the hash values from one signature that are in one or more
of the others.

For example,


sourmash signature subtract file1.sig file2.sig file3.sig -o subtracted.sig

will subtract all of the hashes in `file2.sig` and `file3.sig` from
`file1.sig`, and save the new signature to `subtracted.sig`.

To use `subtract` on signatures calculated with
`-p abund`, you must specify `--flatten`.

Note: `subtract` only creates one output file, with one signature in it.

"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg


def subparser(subparsers):
    subparser = subparsers.add_parser("subtract", description=__doc__, usage=usage)
    subparser.add_argument("signature_from")
    subparser.add_argument("subtraction_sigs", nargs="+")
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        default="-",
        help="output signature to this file (default stdout)",
    )
    subparser.add_argument(
        "--flatten",
        action="store_true",
        help="remove abundance from signatures before subtracting",
    )
    subparser.add_argument(
        "-A",
        "--abundances-from",
        metavar="FILE",
        help="intersect with & take abundances from this signature",
    )
    subparser.add_argument("--set-name", help="set name for output signature")
    add_ksize_arg(subparser)
    add_moltype_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.subtract(args)



"""
Tests for functions in sourmash_args module.
"""

import sys
import os
import pytest
import gzip
import zipfile
import io
import contextlib
import csv
import argparse
import shutil
import json

import sourmash_tst_utils as utils
import sourmash
from sourmash import sourmash_args, manifest
from sourmash.index import LinearIndex
from sourmash.cli.utils import add_ksize_arg

from sourmash.signature import load_signatures_from_json, save_signatures_to_json


def test_save_signatures_api_none():
    # save to sigfile
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    with sourmash_args.SaveSignaturesToLocation(None) as save_sig:
        print(repr(save_sig))
        save_sig.add(ss2)
        save_sig.add(ss47)

    # nothing to test - no output!


def test_save_signatures_to_location_1_sig(runtmp):
    # save to sigfile.sig
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.sig")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_stdout():
    # save to stdout
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    output_capture = io.StringIO()
    with contextlib.redirect_stdout(output_capture):
        with sourmash_args.SaveSignaturesToLocation("-") as save_sig:
            save_sig.add(ss2)
            save_sig.add(ss47)

    output = output_capture.getvalue()

    saved = list(load_signatures_from_json(output))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_sig_is_default(runtmp):
    # save to sigfile.txt
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.txt")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    saved = list(load_signatures_from_json(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_sig_gz(runtmp):
    # save to sigfile.gz
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.sig.gz")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    # can we open as a .gz file?
    with gzip.open(outloc, "r") as fp:
        print(save_sig)
        fp.read()

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_zip(runtmp):
    # save to sigfile.zip
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.zip")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    # can we open as a .zip file?
    with zipfile.ZipFile(outloc, "r") as zf:
        assert list(zf.infolist())

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_zip_bad(runtmp):
    # try saving to bad sigfile.zip
    sig2 = utils.get_test_data("2.fa.sig")
    sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.zip")

    # create bad zip:
    with open(outloc, "w"):
        pass

    # now check for error
    with pytest.raises(ValueError) as exc:
        with sourmash_args.SaveSignaturesToLocation(outloc):
            pass

    assert "cannot be opened as a zip file" in str(exc)


def test_save_signatures_to_location_1_zip_dup(runtmp):
    # save to sigfile.zip
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("foo.zip")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

        # here we have to change the names so the sig content is different;
        # exact duplicates will not be saved, otherwise.
        ss2 = ss2.to_mutable()
        ss2.name = "different name for ss2"
        save_sig.add(ss2)

        ss47 = ss47.to_mutable()
        ss47.name = "different name for ss47"
        save_sig.add(ss47)

    # can we open as a .zip file?
    with zipfile.ZipFile(outloc, "r") as zf:
        assert list(zf.infolist())

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 4


def test_save_signatures_to_location_2_zip_add(runtmp):
    # create sigfile.zip; then, add a new signature.
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    # add only ss2
    outloc = runtmp.output("foo.zip")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)

    # can we open as a .zip file?
    with zipfile.ZipFile(outloc, "r") as zf:
        assert list(zf.infolist())

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert len(saved) == 1

    # now, re-open and add ss47.
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss47)

    # updated file should contain both.
    saved = list(sourmash.load_file_as_signatures(outloc))
    print(saved)
    assert ss47 in saved
    assert ss2 in saved


def test_save_signatures_to_location_2_zip_add_dup(runtmp):
    # create sigfile.zip; then, add a new signature, plus a ~duplicate.
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    # add only ss2
    outloc = runtmp.output("foo.zip")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)

    # can we open as a .zip file?
    with zipfile.ZipFile(outloc, "r") as zf:
        assert list(zf.infolist())

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert len(saved) == 1

    # now, re-open and add ss47, plus a slightly renamed ss2.
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss47)

        # add ss2; here we have to change the names so the sig content is
        # different exact duplicates will not be saved, otherwise.
        import copy

        ss2copy = ss2.to_mutable()
        ss2copy.name = "different name for ss2"
        save_sig.add(ss2copy)

    # updated file should contain all three.
    saved = list(sourmash.load_file_as_signatures(outloc))
    print(len(saved), saved)
    assert ss47 in saved
    assert ss2 in saved
    assert ss2copy in saved


def test_save_signatures_to_location_3_zip_add_fail(runtmp):
    # create sigfile.zip using zipfile, then try to add to it (& fail)
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    sourmash.load_one_signature(sig47, ksize=31)

    # add only ss2, using zipfile API
    outloc = runtmp.output("foo.zip")
    with zipfile.ZipFile(outloc, "x") as zf:
        with zf.open("xyz.sig", "w") as fp:
            save_signatures_to_json([ss2], fp=fp, compression=1)

    # verify it can be loaded, yada yada
    saved = list(sourmash.load_file_as_signatures(outloc))
    print(len(saved), saved)
    assert ss2 in saved

    # now, try to open existing file with SaveSignaturesToLocation...
    with pytest.raises(ValueError) as exc:
        with sourmash_args.SaveSignaturesToLocation(outloc):
            pass

    assert "Cannot add to existing zipfile" in str(exc)


def test_save_signatures_to_location_3_zip_add_with_manifest(runtmp):
    # create sigfile.zip using zipfile, then try to add to it (& fail)
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    # add only ss2, using zipfile API; add manifest manually.
    outloc = runtmp.output("foo.zip")
    with zipfile.ZipFile(outloc, "x") as zf:
        with zf.open("xyz.sig", "w") as fp:
            save_signatures_to_json([ss2], fp=fp, compression=1)

        # make a manifest row...
        row = manifest.CollectionManifest.make_manifest_row(
            ss2, "xyz.sig", include_signature=False
        )

        # construct & save manifest
        mf = manifest.CollectionManifest([row])
        mf_name = "SOURMASH-MANIFEST.csv"

        manifest_fp = io.StringIO()
        mf.write_to_csv(manifest_fp, write_header=True)
        manifest_data = manifest_fp.getvalue().encode("utf-8")

        with zf.open(mf_name, "w") as fp:
            fp.write(manifest_data)

        # fini! made our artisanal hand-crafted zipfile. Now...

    # verify it can be loaded, yada yada
    saved = list(sourmash.load_file_as_signatures(outloc))
    print(len(saved), saved)
    assert ss2 in saved

    # now, try to open existing file with SaveSignaturesToLocation...
    # ...should succeed!
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss47)

    # updated file should contain both
    saved = list(sourmash.load_file_as_signatures(outloc))
    print(len(saved), saved)
    assert ss47 in saved
    assert ss2 in saved


def test_save_signatures_to_location_1_dirout(runtmp):
    # save to sigout/ (directory)
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("sigout/")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    assert os.path.isdir(outloc)

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 2


def test_save_signatures_to_location_1_dirout_bug_2751(runtmp):
    # check for 2x compressed sig files
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("sigout/")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)

    assert os.path.isdir(outloc)
    print(os.listdir(outloc))

    outloc2 = runtmp.output("sigout/09a08691ce52952152f0e866a59f6261.sig.gz")
    with gzip.open(outloc2, "r") as fp:
        data = fp.read()
        print(data)
        _ = json.loads(data)


def test_save_signatures_to_location_1_dirout_duplicate(runtmp):
    # save to sigout/ (directory)
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    outloc = runtmp.output("sigout/")
    with sourmash_args.SaveSignaturesToLocation(outloc) as save_sig:
        print(save_sig)
        save_sig.add(ss2)
        save_sig.add(ss47)
        save_sig.add(ss2)
        save_sig.add(ss47)

    assert os.path.isdir(outloc)

    saved = list(sourmash.load_file_as_signatures(outloc))
    assert ss2 in saved
    assert ss47 in saved
    assert len(saved) == 4


def test_load_empty_zipfile(runtmp):
    outloc = runtmp.output("empty.zip")
    with sourmash_args.SaveSignaturesToLocation(outloc):
        pass

    sigiter = sourmash.load_file_as_signatures(outloc)
    assert list(sigiter) == []


def test_load_many_sigs_empty_file(runtmp):
    # make sure load_many_signatures behaves properly on empty file
    outloc = runtmp.output("empty.sig")
    with open(outloc, "w"):
        pass

    progress = sourmash_args.SignatureLoadingProgress()

    with contextlib.redirect_stderr(io.StringIO()) as errfp:
        with pytest.raises(SystemExit):
            for ss, sigloc in sourmash_args.load_many_signatures([outloc], progress):
                pass

    err = errfp.getvalue()
    print(err)
    assert f"ERROR: Error while reading signatures from '{outloc}'." in err
    assert "(continuing)" not in err


def test_load_many_sigs_empty_file_force(runtmp):
    # make sure load_many_signatures behaves properly on empty file w/force
    outloc = runtmp.output("empty.sig")
    with open(outloc, "w"):
        pass

    progress = sourmash_args.SignatureLoadingProgress()

    with contextlib.redirect_stderr(io.StringIO()) as errfp:
        for ss, sigloc in sourmash_args.load_many_signatures(
            [outloc], progress, force=True
        ):
            pass

    err = errfp.getvalue()
    print(err)
    assert f"ERROR: Error while reading signatures from '{outloc}'." in err
    assert "(continuing)" in err


def test_get_manifest_1():
    # basic get_manifest retrieves a manifest
    sig47 = utils.get_test_data("47.fa.sig")
    idx = sourmash.load_file_as_index(sig47)

    manifest = sourmash_args.get_manifest(idx)
    assert len(manifest) == 1


def test_get_manifest_2_cannot_build():
    # test what happens when get_manifest cannot build manifest
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47)

    idx = LinearIndex([ss47])

    with pytest.raises(SystemExit):
        sourmash_args.get_manifest(idx)


def test_get_manifest_2_cannot_buildno_require():
    # test what happens when get_manifest cannot build manifest
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47)

    idx = LinearIndex([ss47])

    m = sourmash_args.get_manifest(idx, require=False)

    assert m is None


def test_get_manifest_3_build():
    # check that manifest is building
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47)

    class FakeIndex(LinearIndex):
        was_called = 0

        def _signatures_with_internal(self):
            self.was_called = 1
            return [(ss47, "fakeiloc")]

    idx = FakeIndex([sig47])

    assert not idx.was_called
    m = sourmash_args.get_manifest(idx)
    assert idx.was_called

    print(m)
    assert len(m) == 1
    assert m.rows[0]["internal_location"] == "fakeiloc"


def test_get_manifest_3_build_2():
    # check that manifest is building, but only when asked
    sig47 = utils.get_test_data("47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47)

    class FakeIndex(LinearIndex):
        manifest = None
        was_called = 0

        def _signatures_with_internal(self):
            self.was_called = 1
            return [(ss47, "fakeiloc")]

    idx = FakeIndex([sig47])

    assert not idx.was_called
    m = sourmash_args.get_manifest(idx)
    assert idx.was_called

    # now set and ask again, should not be called
    idx.manifest = m
    idx.was_called = 0

    m2 = sourmash_args.get_manifest(idx)
    assert not idx.was_called
    assert m == m2

    # now, force rebuild
    m3 = sourmash_args.get_manifest(idx, rebuild=True)
    assert idx.was_called
    assert m == m3


class FakeArgs:
    picklist = None
    include_db_pattern = None
    exclude_db_pattern = None


def test_pattern_0():
    # test neither --include nor --exclude
    args = FakeArgs()
    args.picklist = None
    args.include_db_pattern = None
    args.exclude_db_pattern = None

    pattern_search = sourmash_args.load_include_exclude_db_patterns(args)
    assert pattern_search is None


def test_pattern_1():
    # test just --include-pattern handling
    args = FakeArgs()
    args.picklist = None
    args.include_db_pattern = "foo"
    args.exclude_db_pattern = None

    pattern_search = sourmash_args.load_include_exclude_db_patterns(args)
    assert pattern_search(["foo", "bar", "baz"])
    assert not pattern_search(["bar", "bif"])


def test_pattern_2():
    # test just --exclude-pattern handling
    args = FakeArgs()
    args.picklist = None
    args.exclude_db_pattern = "foo"
    args.include_db_pattern = None

    pattern_search = sourmash_args.load_include_exclude_db_patterns(args)
    assert not pattern_search(["foo", "bar", "baz"])
    assert pattern_search(["bar", "baz", "bif"])


def test_pattern_3():
    # test with --picklist and --exclude: should fail
    args = FakeArgs()
    args.picklist = True
    args.exclude_db_pattern = "foo"
    args.include_db_pattern = None

    with pytest.raises(SystemExit):
        sourmash_args.load_include_exclude_db_patterns(args)


def test_pattern_4():
    # test with --picklist and --include: should fail
    args = FakeArgs()
    args.picklist = True
    args.include_db_pattern = "foo"
    args.exclude_db_pattern = None

    with pytest.raises(SystemExit):
        sourmash_args.load_include_exclude_db_patterns(args)


def test_pattern_5():
    # test with --include and --exclude: should fail
    args = FakeArgs()
    args.picklist = None
    args.exclude_db_pattern = "foo"
    args.include_db_pattern = "bar"

    with pytest.raises(SystemExit):
        sourmash_args.load_include_exclude_db_patterns(args)


def test_fileinput_csv_1_plain():
    # test basic CSV input

    testfile = utils.get_test_data("tax/test.taxonomy.csv")

    with sourmash_args.FileInputCSV(testfile) as r:
        rows = list(r)
        assert len(rows) == 6


def test_fileinput_csv_1_no_such_file(runtmp):
    # test fail to load file

    noexistfile = runtmp.output("does-not-exist.csv")

    with pytest.raises(FileNotFoundError):
        with sourmash_args.FileInputCSV(noexistfile):
            pass


def test_fileinput_csv_2_gz(runtmp):
    # test basic CSV input from gz file

    testfile = utils.get_test_data("tax/test.taxonomy.csv")
    gzfile = runtmp.output("test.csv.gz")

    with gzip.open(gzfile, "wt") as outfp:
        with open(testfile, newline="") as infp:
            outfp.write(infp.read())

    with sourmash_args.FileInputCSV(gzfile) as r:
        rows = list(r)
        assert len(rows) == 6


def test_fileinput_csv_2_gz_not_csv(runtmp):
    # test basic CSV input from gz file that's not CSV - works

    gzfile = runtmp.output("test.csv.gz")

    with gzip.open(gzfile, "wt") as outfp:
        outfp.write("hello world!")

    with sourmash_args.FileInputCSV(gzfile) as r:
        assert r.fieldnames == ["hello world!"]


def test_fileinput_csv_2_gz_bad_version_header(runtmp):
    # test basic CSV input from gz file with bad version header
    # currently this works; not clear to me how it should fail :grin:

    gzfile = runtmp.output("test.csv.gz")

    with gzip.open(gzfile, "wt") as outfp:
        outfp.write("# excelsior\nhello world!")

    with sourmash_args.FileInputCSV(gzfile) as r:
        assert r.fieldnames == ["hello world!"]
        print(r.version_info)
        assert r.version_info == ["excelsior"]


def test_fileinput_csv_2_zip(runtmp):
    # test CSV input from zip file, with component filename

    testfile = utils.get_test_data("tax/test.taxonomy.csv")
    zf_file = runtmp.output("test.zip")

    with zipfile.ZipFile(zf_file, "w") as outzip:
        with open(testfile, "rb") as infp:
            with outzip.open("XYZ.csv", "w") as outfp:
                outfp.write(infp.read())

    with sourmash_args.FileInputCSV(zf_file, default_csv_name="XYZ.csv") as r:
        rows = list(r)
        assert len(rows) == 6
        print(rows)


def test_fileinput_csv_3_load_manifest():
    # test loading a manifest from a zipfile collection, using
    # FileInputCSV.
    testfile = utils.get_test_data("prot/all.zip")

    with sourmash_args.FileInputCSV(
        testfile, default_csv_name="SOURMASH-MANIFEST.csv"
    ) as r:
        rows = list(r)
        assert len(rows) == 8

        assert r.version_info == ["SOURMASH-MANIFEST-VERSION", "1.0"]


def test_fileinput_csv_3_load_manifest_no_default():
    # test loading a manifest from a zipfile collection, using
    # FileInputCSV, but with no default_csv_name - should fail
    testfile = utils.get_test_data("prot/all.zip")

    with pytest.raises(csv.Error):
        with sourmash_args.FileInputCSV(testfile) as r:
            print(r.fieldnames)


def test_fileinput_csv_3_load_manifest_zipfile_obj():
    # test loading a manifest from an open zipfile obj, using
    # FileInputCSV.
    testfile = utils.get_test_data("prot/all.zip")

    with zipfile.ZipFile(testfile, "r") as zf:
        with sourmash_args.FileInputCSV(
            testfile, default_csv_name="SOURMASH-MANIFEST.csv", zipfile_obj=zf
        ) as r:
            rows = list(r)
            assert len(rows) == 8

            assert r.version_info == ["SOURMASH-MANIFEST-VERSION", "1.0"]


def test_fileinput_csv_3_load_manifest_zipfile_obj_no_defualt():
    # test loading a manifest from an open zipfile obj, using
    # FileInputCSV, but with no default csv name => should fail.
    testfile = utils.get_test_data("prot/all.zip")

    with zipfile.ZipFile(testfile, "r") as zf:
        with pytest.raises(ValueError):
            with sourmash_args.FileInputCSV(testfile, zipfile_obj=zf):
                pass


def test_fileoutput_csv_1(runtmp):
    # test basic behavior
    outfile = runtmp.output("xxx.csv")

    with sourmash_args.FileOutputCSV(outfile) as fp:
        w = csv.writer(fp)
        w.writerow(["a", "b", "c"])
        w.writerow(["x", "y", "z"])

    with open(outfile, newline="") as fp:
        r = csv.DictReader(fp)
        rows = list(r)
        assert len(rows) == 1
        row = rows[0]
        assert row["a"] == "x"
        assert row["b"] == "y"
        assert row["c"] == "z"


def test_fileoutput_csv_1_gz(runtmp):
    # test basic behavior => gz
    outfile = runtmp.output("xxx.csv.gz")

    with sourmash_args.FileOutputCSV(outfile) as fp:
        w = csv.writer(fp)
        w.writerow(["a", "b", "c"])
        w.writerow(["x", "y", "z"])

    with gzip.open(outfile, "rt") as fp:
        r = csv.DictReader(fp)
        rows = list(r)
        assert len(rows) == 1
        row = rows[0]
        assert row["a"] == "x"
        assert row["b"] == "y"
        assert row["c"] == "z"


def test_fileoutput_csv_2_stdout():
    # test '-' and 'None' go to sys.stdout

    with sourmash_args.FileOutputCSV("-") as fp:
        assert fp == sys.stdout

    with sourmash_args.FileOutputCSV(None) as fp:
        assert fp == sys.stdout


def test_add_ksize_arg_no_default():
    # test behavior of cli.utils.add_ksize_arg
    p = argparse.ArgumentParser()
    add_ksize_arg(p)
    args = p.parse_args()
    assert args.ksize is None


def test_add_ksize_arg_no_default_specify():
    # test behavior of cli.utils.add_ksize_arg
    p = argparse.ArgumentParser()
    add_ksize_arg(p)
    args = p.parse_args(["-k", "21"])
    assert args.ksize == 21


def test_add_ksize_arg_default_31():
    # test behavior of cli.utils.add_ksize_arg
    p = argparse.ArgumentParser()
    add_ksize_arg(p, default=31)
    args = p.parse_args()
    assert args.ksize == 31


def test_add_ksize_arg_default_31_specify():
    # test behavior of cli.utils.add_ksize_arg
    p = argparse.ArgumentParser()
    add_ksize_arg(p, default=31)
    args = p.parse_args(["-k", "21"])
    assert args.ksize == 21


def test_bug_2370(runtmp):
    # bug - manifest loading code does not catch gzip.BadGzipFile
    sigfile = utils.get_test_data("63.fa.sig")

    # copy sigfile over to a .gz file without compressing it -
    shutil.copyfile(sigfile, runtmp.output("not_really_gzipped.gz"))

    # try running sourmash_args.load_file_as_index
    # runtmp.sourmash('sig', 'describe', runtmp.output('not_really_gzipped.gz'))
    sourmash_args.load_file_as_index(runtmp.output("not_really_gzipped.gz"))


def test_load_one_signature_1(runtmp):
    # test the sourmash_args.load_one_signature function
    sigfile = utils.get_test_data("63.fa.sig.zip")

    ss = sourmash_args.load_one_signature(sigfile, ksize=31)
    assert ss.name.startswith("NC_011663.1 ")


def test_load_one_signature_2_fail(runtmp):
    # test the sourmash_args.load_one_signature function on failure - no sig
    sigfile = utils.get_test_data("63.fa.sig.zip")

    with pytest.raises(ValueError) as exc:
        sourmash_args.load_one_signature(sigfile, ksize=21)

    assert "expected exactly one." in str(exc)


def test_load_one_signature_3_fail(runtmp):
    # test the sourmash_args.load_one_signature function on failure - many sigs
    sigfile = utils.get_test_data("prot/all.zip")

    with pytest.raises(ValueError) as exc:
        sourmash_args.load_one_signature(sigfile)

    assert "more than one signature" in str(exc)
